function [S,L_k,W] = goka_b(A,p,k,reorth)
% GOKA_B Golub-Kahan bidiagonalization.
%
% L_k = goka_b(A,p,k,reorth)
% [S,L_k,W] = goka_b(A,p,k,reorth)
%
% Performs k steps of the Golub-Kahan bidiagonalization process with
% starting vector p, producing a lower bidiagonal matrix
%           [b_11               ]
%           [b_21 b_22          ]
%     L_k = [     b_32 .        ]
%           [          . b_kk   ]
%           [            b_k+1,k]
% such that
%     A*W = S*L_k ,
% where S and W consist of the left and right bidiagonalization vectors.
%
% Reorthogonalization is controlled by means of reorth:
%    reorth = 0 : no reorthogonalization,
%    reorth = 1 : reorthogonalization by means of MGS,
%    reorth = 2 : Householder-reorthogonalization.
% No reorthogonalization is assumed if reorth is not specified.

% Reference: G. H. Golub & C. F. Van Loan, "Matrix Computations",
% 3. Ed., Johns Hopkins, 1996.  Section 9.3.4.
% Referred to as "bidiag1" by Paige and Saunders.

% Per Christian Hansen, IMM, April 8, 2001.
% Modified by Martin Plesinger, TUL & ASCR, April 2009.

% Initialization.
if (k<1), error('Number of steps k must be positive'), end
if (nargin < 4), reorth = 0; end
if (reorth < 0 | reorth > 2), error('Illegal reorth'), end
if (nargout==2), error('Not enough output arguments'), end
[m,n] = size(A);
L_k = sparse(k+1,k);
if (nargout>1 | reorth==1)
  S = zeros(m,k); W = zeros(n,k); SW = 1;
else
  SW = 0;
end
if (reorth==2)
  if (k>=n), error('No. of iterations must satisfy k < n'), end
  HHS = zeros(m,k); HHW = zeros(n,k);
  HHalpha = zeros(1,k); HHbeta = HHalpha;
end

% Prepare for Golub-Kahan iteration.
w = zeros(n,1);
beta = norm(p);
if (beta==0), error('Starting vector must be nonzero'), end
if (reorth==2)
  [beta,HHbeta(1),HHS(:,1)] = gen_hh(p);
end
s = p/beta;
if (SW), S(:,1) = s; end

% Perform Golub-Kahan bidiagonalization with/without reorthogonalization.
for i = 1:k

  r = A'*s - beta*w;
  if (reorth==0)
    alpha = norm(r); w = r/alpha;
  elseif (reorth==1)
    for j=1:i-1, r = r - (W(:,j)'*r)*W(:,j); end
    alpha = norm(r); w = r/alpha;
  else
    for j=1:i-1
      r(j:n) = app_hh(r(j:n),HHalpha(j),HHW(j:n,j));
    end
    [alpha,HHalpha(i),HHW(i:n,i)] = gen_hh(r(i:n));
    w = zeros(n,1); w(i) = 1;
    for j=i:-1:1
      w(j:n) = app_hh(w(j:n),HHalpha(j),HHW(j:n,j));
    end
  end
  L_k(i,i) = alpha; if (SW), W(:,i) = w; end

  p = A*w - alpha*s;
  if (reorth==0)
    beta = norm(p); s = p/beta;
  elseif (reorth==1)
    for j=1:i, p = p - (S(:,j)'*p)*S(:,j); end
    beta = norm(p); s = p/beta;
  else
    for j=1:i
      p(j:m) = app_hh(p(j:m),HHbeta(j),HHS(j:m,j));
    end
    [beta,HHbeta(i+1),HHS(i+1:m,i+1)] = gen_hh(p(i+1:m));
    s = zeros(m,1); s(i+1) = 1;
    for j=i+1:-1:1
      s(j:m) = app_hh(s(j:m),HHbeta(j),HHS(j:m,j));
    end
  end
  L_k(i+1,i) = beta; if (SW), S(:,i+1) = s; end

end

if (nargout==1), S = L_k; end